/* panatest.c */

/**************************************************************************+
*
*  PANAMA high-performance reference C-code, based on the description in 
*  the paper 'Fast Hashing and Stream Encryption with PANAMA', presented 
*  at the Fast Software Encryption Workshop, Paris, 1998, see "Fast 
*  Software Encryption - 5th International Workshop, FSE'98", edited by 
*  Serge Vaudenay, LNCS-1372, Springer-Verlag, 1998, pp 60-74, also 
*  available on-line at http://standard.pictel.com/ftp/research/security
*
*  Algorithm design by Joan Daemen and Craig Clapp
*
*  panatest.c  -  Test harness for Panama stream/hash module, consistent 
*                 with exportable version - panama_x.c
*
*
*  History:
*
*  29-Oct-98  Craig Clapp  Implemention for Dr. Dobbs, Dec. 1998 issue, 
*                          based on earlier performance-benchmark code.
*
*
*  Notes:  This code is supplied for the purposes of evaluating the 
*          performance of the Panama stream/hash module and as a 
*          reference implementation for generating test vectors for 
*          compatibility / interoperability verification.
*
*
+**************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "panama.h"


/**************************************************************************+
*             Test-harness for the Panama cryptographic module             *
+**************************************************************************/


/**************************************************************************+
*
*  pan_hash_test() - Prints the 256-bit hash of the array of bytes given 
*                    by text_array, where the size of the message (in bits, 
*                    which need not be a multiple of eight) can be specified. 
*                    Optionally, a sequence of hashes can be printed for 
*                    progressively truncated versions of the message.
*                    If no array is given (text_array = NULL) then a 
*                    reference set of hashes for sequential length strings 
*                    of all zeros and all ones is printed. Note that hashes 
*                    for the all-zeros and all-ones strings cannot be used 
*                    to validate endian compatibility since these strings 
*                    are endian invariant.
*
+**************************************************************************/

void pan_hash_test(unsigned char *text_array, long min_bit_length, long max_bit_length)
{
    long i, bits;

    UINT32 hashcode[PAN_STAGE_SIZE];

    UINT32 *hashtext = (UINT32*)calloc((max_bit_length + WORDLENGTH - 1) / WORDLENGTH, sizeof(UINT32));
    assert(hashtext != NULL); /* abort if array exceeds available space */


    if (text_array == NULL)
    {
        printf("\n                  Panama hash function test vectors\n");
        printf(  "                  ---------------------------------\n");

        /* perform hashing on various length bit-strings of zeros and ones */

        printf("\n256-bit hash results for a string of zeros of various lengths (in bits) are:\n\n");
        printf("  No.  255  224 223  192 191  160 159  128 127   96 95    64 63    32 31     0\n");
        printf(" bits  |      | |      | |      | |      | |      | |      | |      | |      |");
        for (bits = min_bit_length; bits <= max_bit_length; bits++)
        {
            pan_hash(hashtext, hashcode, bits);

            printf("\n%5ld  ", bits);
            for (i = PAN_STAGE_SIZE-1; i >= 0; i--)
            {
                printf("%08lx ", (ULONG)hashcode[i]);
            }
        }
        printf("\n\n");

        for (i = 0; i < (max_bit_length + WORDLENGTH - 1) / WORDLENGTH; i++)
        {
            hashtext[i] = ~0L;
        }

        printf("\n256-bit hash results for a string of ones of various lengths (in bits) are:\n\n");
        printf("  No.  255  224 223  192 191  160 159  128 127   96 95    64 63    32 31     0\n");
        printf(" bits  |      | |      | |      | |      | |      | |      | |      | |      |");
        for (bits = min_bit_length; bits <= max_bit_length; bits++)
        {
            pan_hash(hashtext, hashcode, bits);

            printf("\n%5ld  ", bits);
            for (i = PAN_STAGE_SIZE-1; i >= 0; i--)
            {
                printf("%08lx ", (ULONG)hashcode[i]);
            }
        }
        printf("\n\n");
    }
    else
    {
        /* This test case exercises byte-to-UINT32 packing conventions (endianness) */
        /* These test vectors [arbitrarily] define Panama to use little-endian packing conventions */

        /* convert text_array to consistent (little-endian) representation as UINT32 array */
        /* for little-endian machines with a native 32-bit data type this conversion would 
           normally be done by casting the byte array to an array of 32-bit words */
        /* here we implement explicit byte-to-UINT32 packing in order to produce definitive 
           test vectors regardless of the target machine's endianness or word size */

        for (i = 0; i < (max_bit_length + WORDLENGTH - 1) / WORDLENGTH; i++)
        {
            hashtext[i] = (((UINT32)text_array[4*i+3]) << 24) + (((UINT32)text_array[4*i+2]) << 16) 
                        + (((UINT32)text_array[4*i+1]) <<  8) +  ((UINT32)text_array[4*i]);
        }


        if (min_bit_length == max_bit_length)
        {
            printf("\n256-bit hash for the ASCII string:\n");
            printf(" \42%s\42\ncropped to %ld bits is:\n\n", text_array, (ULONG)max_bit_length);
        }
        else
        {
            printf("\n256-bit hash results for the ASCII string:\n");
            printf(" \42%s\42\ncropped to various lengths (in bits) are:\n\n", text_array);
        }

        printf("  No.  255  224 223  192 191  160 159  128 127   96 95    64 63    32 31     0\n");
        printf(" bits  |      | |      | |      | |      | |      | |      | |      | |      |");

        for (bits = min_bit_length; bits <= max_bit_length; bits++)
        {
            pan_hash(hashtext, hashcode, bits);

            printf("\n%5ld  ", bits);
            for (i = PAN_STAGE_SIZE-1; i >= 0; i--)
            {
                printf("%08lx ", (ULONG)hashcode[i]);
            }
        }
        printf("\n\n");
    }

    free((void*)hashtext);
}


/**************************************************************************+
*
*  pan_speed_test() - Performs a series of hashing and encryption passes 
*                     a over large array and reports their measured speeds.
*
+**************************************************************************/

#include <time.h>  /* needed for performance measurements */

#ifndef CLOCKS_PER_SEC     /* may occur if compiler lacks support for timing */
#define CLOCKS_PER_SEC -1  /* dummy value, used as a flag */
#endif


#define MESSAGE_LENGTH 25000    /* number of 32-bit words in test array */
#define PASSES 1000             /* default number of passes for timing measurement */

#define MESSAGE_BYTES  (1L * MESSAGE_LENGTH * WORDLENGTH / 8)


void pan_speed_test(long test_passes)
{
    long i;
    clock_t start, finish;
    double  duration, rate, bytes_processed;

    PAN_BUFFER buffer;
    PAN_STATE state;
    UINT32 hashcode[PAN_STAGE_SIZE];

    UINT32 *plaintext  = (UINT32*)calloc(MESSAGE_LENGTH, sizeof(UINT32));
    UINT32 *ciphertext = (UINT32*)calloc(MESSAGE_LENGTH, sizeof(UINT32));

    assert(ciphertext != NULL); /* abort if test arrays exceed available space */


    printf("\n                Panama encryption/hashing speed tests\n");
    printf(  "                -------------------------------------\n");

    /* run large-block performance measurements */

    printf("\nEach test consists of %ld passes over a buffer of length %ld bytes\n", test_passes, (ULONG)MESSAGE_BYTES);

    bytes_processed = (double)1.0 * test_passes * MESSAGE_BYTES;
    printf("%0.0f bytes in total are processed in each test\n\n", bytes_processed);

    if (clock() == -1)
    {
        printf("clock() function not supported by C-compiler\n");
    }

    if (CLOCKS_PER_SEC == -1)
    {
        printf("CLOCKS_PER_SEC parameter not supported by C-compiler\n");
    }

    if (clock() == -1 || CLOCKS_PER_SEC == -1)
    {
        printf("Automatic timing measurements not available on this platform\n");
        printf("Timing measurements should be performed manually (e.g. by stopwatch)\n\n");
    }

    printf("Starting memory-to-memory encryption speed test ...\n");

    /* repeat simulated encryption several times on large block for speed measurement */
    /* first test uses distinct source and destination arrays */
    start = clock();
    for (i = 0; i < test_passes; i++)
    {
        pan_reset(&buffer, &state);
        pan_pull(plaintext, ciphertext, MESSAGE_LENGTH / PAN_STAGE_SIZE, &buffer, &state);
    }
    finish = clock();

    printf("Memory-to-memory encryption speed test completed, %0.0f bytes processed\n\n", bytes_processed);

    if (clock() != -1 && CLOCKS_PER_SEC != -1)
    {
        duration = (double)(finish - start) / CLOCKS_PER_SEC;
        printf("Time taken for memory-to-memory encryption = %0.3f seconds\n", duration);
        rate = (double)1.0 * test_passes * MESSAGE_BYTES / duration / 1e6;
        printf("Encryption speed = %0.3f MBytes/s\n\n", rate);
    }

    printf("Tail-end of randomized buffer is:\n");
    printf("     n-7      n-6      n-5      n-4      n-3      n-2      n-1       n\n   ");
    for (i = PAN_STAGE_SIZE; i > 0; i--)
    {
        printf("%08lx ", (ULONG)ciphertext[MESSAGE_LENGTH - i]);
    }
    printf("\n\n");


    printf("Starting in-place encryption speed test ...\n");

    /* repeat simulated encryption several times on large block for speed measurement */
    /* this time do it in-place to minimize cache misses */
    start = clock();
    for (i = 0; i < test_passes; i++)
    {
        pan_reset(&buffer, &state);
        pan_pull(plaintext, plaintext, MESSAGE_LENGTH / PAN_STAGE_SIZE, &buffer, &state);
    }
    finish = clock();

    printf("In-place encryption completed, %0.0f bytes processed\n\n", bytes_processed);

    if (clock() != -1 && CLOCKS_PER_SEC != -1)
    {
        duration = (double)(finish - start) / CLOCKS_PER_SEC;
        printf("Time taken for in-place memory-to-memory encryption = %0.3f seconds\n", duration);
        rate = (double)1.0 * test_passes * MESSAGE_BYTES / duration / 1e6;
        printf("In-place encryption speed = %0.3f MBytes/s\n\n", rate);
    }


    printf("Starting hash speed test ...\n");

    /* repeat hash several times on large block for speed measurement */
    start = clock();
    for (i = 0; i < test_passes; i++)
    {
        pan_hash(ciphertext, hashcode, 1L * MESSAGE_LENGTH * WORDLENGTH);
    }
    finish = clock();

    printf("Hash speed test completed, %0.0f bytes processed\n\n", bytes_processed);

    if (clock() != -1 && CLOCKS_PER_SEC != -1)
    {
        duration = (double)(finish - start) / CLOCKS_PER_SEC;
        printf("Time taken for hashing = %0.3f seconds\n", duration);
        rate = (double)1.0 * test_passes * MESSAGE_BYTES / duration / 1e6;
        printf("Hashing speed = %0.3f MBytes/s\n\n", rate);
    }

    printf("256-bit hash result is:\n");
    printf("   255  224 223  192 191  160 159  128 127   96 95    64 63    32 31     0\n");
    printf("   |      | |      | |      | |      | |      | |      | |      | |      |\n   ");
    for (i = PAN_STAGE_SIZE-1; i >= 0; i--)
    {
        printf("%08lx ", (ULONG)hashcode[i]);
    }
    printf("\n\n");

    free((void*)plaintext);
    free((void*)ciphertext);
}


/**************************************************************************+
*
*  panatest.c  -  main()
*
*  The main routine can be called with up to three command-line arguments:
*
*  The first (optional) argument specifies the number of times to perform 
*  encryption and hashing  over the test array. For performance measurements 
*  this number should be chosen large enough to swamp the granularity of the 
*  real-time clock. (e.g. a PC's real-time clock has a granularity of >50ms 
*  so each test case should run for around 5 secs. to get reasonable 
*  accuracy). Setting this argument to 0 (zero) causes this part of the 
*  program to be  skipped.
*
*  The second (optional) argument gives a text string to be hashed. Use 
*  the target platform's standard command-line conventions if spaces or 
*  control characters are to be included (e.g. enclose the string in double 
*  quotes for MS-DOS command line).
*
*  The third (optional) argument specifies the length (in bits) that the 
*  text string is to be truncated to before the hash is generated. This 
*  number cannot be longer than the actual length of the string in bits.
*
*  If all arguments are omitted the program will generate performance 
*  figures for it's default number of passes, and will generate a reference 
*  set of test vectors. To generate test vectors only, invoke panatest 
*  with a single argument of 0 (zero).
*
*  Examples:
*
*   panatest        - Print performance figures and reference test vectors
*   panatest 0      - Print reference test vectors only
*   panatest 2500   - Print performance figures for 2500 runs of the 
*                     performance test
*
*   panatest 0 "Hello world"       - Print the hash for Hello world
*   panatest 100 "Hello world" 39  - Run 100 passes of the performance test 
*                                    and print the hash for the 39-bit 
*                                    string had by truncating Hello world.
*
+**************************************************************************/

int main(int argc, char *argv[])
{
    long bits;
    long test_passes = PASSES;

    unsigned char text_array[] = "The quick brown fox jumps over the lazy dog";


    if (argc > 1)
    {
        test_passes = atol(argv[1]);
    }

    if (test_passes > 0)
    {
        pan_speed_test(test_passes);
    }


    if (argc > 2)
    {
        bits = 8 * strlen(argv[2]);

        if (argc > 3 && atol(argv[3]) < bits)
        {
            bits = atol(argv[3]);
        }

        pan_hash_test((unsigned char*)(argv[2]), bits, bits);
    }
    else
    {
        if (argc == 1 || argc == 2 && test_passes == 0)
        {
            pan_hash_test(NULL, 0, 320);
            pan_hash_test((unsigned char*)text_array, 8, 8 * (sizeof(text_array) - 1));
        }
    }

    return(0);
}
